"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fse = __importStar(require("fs-extra"));
const pixelmatch_1 = __importDefault(require("pixelmatch"));
const path_1 = __importDefault(require("path"));
const pngjs_1 = require("pngjs");
const assert_1 = __importDefault(require("assert"));
const is_docker_1 = __importDefault(require("is-docker"));
const util_1 = require("util");
/**
 * Standard viewport sizes.
 */
class ViewPortSize {
}
exports.ViewPortSize = ViewPortSize;
/**
 * The default browser viewport size.
 */
ViewPortSize.default = { width: 1440, height: 900 };
/**
 * Image comparison result type.
 */
var ScreenCheckResultType;
(function (ScreenCheckResultType) {
    /**
     * No base image located.
     */
    ScreenCheckResultType["NO_BASE_IMAGE"] = "NO_BASE_IMAGE";
    /**
     * The images are alike.
     */
    ScreenCheckResultType["SAME"] = "SAME";
    /**
     * The images are different.
     */
    ScreenCheckResultType["DIFFERENT"] = "DIFFERENT";
})(ScreenCheckResultType = exports.ScreenCheckResultType || (exports.ScreenCheckResultType = {}));
/**
 * Encapsulates a screen capture comparison. This is the type returned by the CLI method `screencheck`.
 */
class ScreenCheckResult {
    constructor(result = ScreenCheckResultType.SAME, data, referenceData, pixelCount) {
        this.result = result;
        this.data = data;
        this.referenceData = referenceData;
        this.pixelCount = pixelCount || 0;
    }
    /**
     * A PNG data buffer containing pixel data corresponding to the difference between data and reference data.
     *
     * Use this method to get a render ready diff image.
     */
    async getDiff() {
        if (this.referenceData) {
            return (await ScreenCheck.compareImages(this.referenceData, this.data)).diffImage.data;
        }
    }
    /**
     * The string representation of the result; e.g. "SAME", "DIFFERENT", "NO_BASE_IMAGE"
     */
    toString() {
        return this.result.toString();
    }
    /**
     * true if result is [[ScreenCheckResultType.SAME]]
     */
    isSame() {
        return this.result === ScreenCheckResultType.SAME;
    }
    /**
     * true if result is [[ScreenCheckResultType.DIFFERENT]]
     */
    isDifferent() {
        return this.result === ScreenCheckResultType.DIFFERENT;
    }
    /**
     * Asserts result is [[ScreenCheckResultType.SAME]]
     *
     * @throws AssertionError
     */
    okay() {
        assert_1.default.equal(this.result, ScreenCheckResultType.SAME, "the screenshot does not match the reference screenshot");
    }
}
exports.ScreenCheckResult = ScreenCheckResult;
const pendingExports = [];
function entryPoint() {
    return function (target, key, descriptor) {
        const init = descriptor.value;
        descriptor.value = function (taiko) {
            init(taiko);
            while (pendingExports.length) {
                let { name, value } = pendingExports.pop();
                // @ts-ignore
                taiko[name] = value;
            }
        };
    };
}
function exportForPlugin(name) {
    return async function (target, key, descriptor) {
        pendingExports.push({ name, value: descriptor.value });
    };
}
/**
 * <strong>Decorator</strong>
 *
 * Caches the return value of a function.
 *
 * <strong>Warning:</strong> Does not consider parameters in cache key. Not recommended for memoizing functions with arity greater than zero.
 * @param duration ms duration for return value TTL
 */
function cached(duration = NaN) {
    return function (target, key, descriptor) {
        const func = descriptor.value;
        descriptor.value = function () {
            if (descriptor.value._has_run && descriptor.value._expires <= Date.now()) {
                return descriptor.value._return;
            }
            descriptor.value._return = func(...arguments);
            descriptor.value._expires = isNaN(duration) ? 0 : Date.now() + Math.ceil(duration);
            descriptor.value._has_run = true;
            return descriptor.value._return;
        };
        descriptor.value.invalidate = () => {
            delete descriptor.value._return;
            delete descriptor.value._expires;
            delete descriptor.value._has_run;
        };
    };
}
function cacheInvalidate(method) {
    return function (target, propertyKey) {
        if (delete target[propertyKey]) {
            let value;
            Object.defineProperty(target, propertyKey, {
                get: () => value,
                set: (newValue) => {
                    if (value !== newValue) {
                        if (target[method].invalidate) {
                            target[method].invalidate();
                        }
                    }
                    value = newValue;
                },
                enumerable: true,
                configurable: true
            });
        }
    };
}
/**
 * Setup options for screencheck. Allows optional configuration of base directory, run id, reference run id, and filename generator function.
 */
class ScreenCheckSetup {
}
exports.ScreenCheckSetup = ScreenCheckSetup;
class ScreenCheck {
    /**
     * The entrypoint for the taiko plugin. Should be called explicitly only when imported programmatically.
     * @entrypoint
     * @param taiko taiko exported API
     */
    static init(taiko) {
        /** hooks */
        ScreenCheck._openBrowser = taiko.openBrowser;
        ScreenCheck.filenameGenerator = ScreenCheck.defaultFilenameGenerator;
        ScreenCheck.taiko = taiko;
        ScreenCheck.baseDir = "";
        ScreenCheck.runId = "";
        ScreenCheck.refRunId = "";
        ScreenCheck.setup();
    }
    /**
     * opens the browser, ensuring renderer pixel ratio is 1:1 to display device
     * @param options taiko openBrowser options
     * @param useOriginalCall use default arguments when launching chromium, otherwise forces 1:1 pixel ratio (default)
     */
    static async openBrowser(options = {}, useOriginalCall = false) {
        if (useOriginalCall) {
            await ScreenCheck._openBrowser(options);
        }
        else {
            options.args = [...(options.args || []),
                '--disable-gpu',
                '--high-dpi-support=1',
                '--device-scale-factor=1',
                '--force-device-scale-factor=1',
                ...(ScreenCheck.isContainerEnvironment() ? [
                    '--headless',
                    '--no-sandbox'
                ] : [])
            ];
            await ScreenCheck._openBrowser(options);
            await ScreenCheck.taiko.setViewPort(ViewPortSize.default);
        }
    }
    /**
     * Configures screencheck.
     * @CLI `screencheckSetup()`
     * @param options screencheck configuration
     */
    static setup(options) {
        ScreenCheck.baseDir = options && options.baseDir || ScreenCheck.baseDir || process.cwd();
        ScreenCheck.runId = options && options.runId ? options.runId : ScreenCheck.runId || ScreenCheck.nextRunId();
        ScreenCheck.refRunId = options && options.refRunId ? options.refRunId : ScreenCheck.refRunId || ScreenCheck.latestRunId();
        ScreenCheck.filenameGenerator = options && options.filenameGenerator ? options.filenameGenerator : ScreenCheck.filenameGenerator;
        ScreenCheck.isSetup = true;
        return {
            baseDir: ScreenCheck.baseDir,
            runId: ScreenCheck.runId,
            refRunId: ScreenCheck.refRunId,
            filenameGenerator: ScreenCheck.filenameGenerator
        };
    }
    /**
     * Determines the latest run index. Useful when using default setup data.
     */
    static detectLatestAutoRunIdIndex() {
        let current = fse.readdirSync(ScreenCheck.baseDir)
            .filter(name => name.match(/^\d+\./))
            .filter(name => fse.lstatSync(name).isDirectory)
            .sort()
            .map(name => parseInt(name.split(".")[0]))
            .reverse()[0]
            || 0;
        return current;
    }
    /**
     * Latest run. This value determines the relative directory from which to load reference data when default setup is used.
     */
    static latestRunId() {
        const current = ScreenCheck.detectLatestAutoRunIdIndex();
        return `${current.toString().padStart(4, "0")}.auto`;
    }
    /**
     * Current run. This value determines the relative directory to store data when default setup is used.
     */
    static nextRunId() {
        const current = ScreenCheck.detectLatestAutoRunIdIndex();
        let next = current + 1;
        return `${next.toString().padStart(4, "0")}.auto`;
    }
    /**
     * The screencheck method that is intended to replace the taiko builtin screenshot method. This method is exported to the taiko CLI
     *
     * @CLI `screencheck()`
     * @param options taiko screenshot options.
     * @param args taiko search elements
     */
    static async screencheck(options, ...args) {
        if (!ScreenCheck.isSetup)
            ScreenCheck.setup();
        options = options || {};
        if (options.path && path_1.default.isAbsolute(options.path)) {
            throw "options.path cannot be absolute. Please specify a directory relative to .baseDir";
        }
        const relativeFilename = options.path || await ScreenCheck.filenameGenerator(options, ...args);
        const referenceImage = await ScreenCheck.getReferenceImagePath(relativeFilename);
        options.path = `${ScreenCheck.getRunDir()}/${relativeFilename}`;
        await fse.mkdirp(ScreenCheck.getRunDir() + "/");
        await ScreenCheck.taiko.screenshot(options, ...args);
        if (referenceImage) {
            const diff = await ScreenCheck.compareImages(referenceImage, options.path);
            if (diff.missmatching > 0) {
                return new ScreenCheckResult(ScreenCheckResultType.DIFFERENT, fse.readFileSync(options.path), fse.readFileSync(referenceImage), diff.missmatching);
            }
            else {
                return new ScreenCheckResult(ScreenCheckResultType.SAME, fse.readFileSync(options.path));
            }
        }
        else {
            return new ScreenCheckResult(ScreenCheckResultType.NO_BASE_IMAGE, fse.readFileSync(options.path));
        }
    }
    /**
     * Returns the absolute path of the reference image corresponding to the given current image relative path.
     * @param current the path of the new image for which a reference image is required
     */
    static async getReferenceImagePath(current) {
        const absolutePath = path_1.default.join(ScreenCheck.getRefDir(), current);
        if (await fse.pathExists(absolutePath)) {
            return absolutePath;
        }
        return undefined;
    }
    /**
     * Compares PNG data and returns a diff PNG (with pngjs) and count of missmatching pixels.
     * @param left reference PNG data buffer
     * @param right new PNG data buffer
     */
    static async compareImages(left, right) {
        const leftImage = pngjs_1.PNG.sync.read(util_1.isBuffer(left) ? left : fse.readFileSync(left));
        const rightImage = pngjs_1.PNG.sync.read(util_1.isBuffer(right) ? right : fse.readFileSync(right));
        const { width, height } = leftImage;
        const diffImage = new pngjs_1.PNG({ width, height });
        const missmatching = pixelmatch_1.default(leftImage.data, rightImage.data, diffImage.data, width, height);
        return {
            missmatching,
            diffImage
        };
    }
    /**
     * Removes structure from a file path to flatten it to a filename.
     * e.g. /foo/bar/buzz.zip becomes foo-bar-buzz-zip
     * @param path a file path to simplify
     */
    static simplifyPath(path) {
        //throw new Error("Method not implemented.")
        return path.replace(/[^a-z0-9_-]+/gi, "-").replace(/^-+/g, "").replace(/-+$/, "");
    }
    /**
     * Generates a name for a unique name for a screenshot.
     * @param taiko an instance of taiko
     * @param fullPage if true is capture of whole page
     * @param args taiko screenshot search elements
     */
    static async generateName(taiko, fullPage = false, ...args) {
        const crypto = require('crypto');
        const url = await taiko.currentURL();
        const urlPart = ScreenCheck.simplifyPath(url.replace(/^\w*\:/gi, ""));
        return `${urlPart}${(args.length ? "-" : "") + args.map(arg => crypto.createHash('md5').update(arg.description).digest("hex").match(/.{6}/)).join("-")}${fullPage ? '-fullpage' : ''}`;
    }
    /**
     * Detect if execution is inside a container environment.
     *
     * @remarks Exposed as isContainerEnvironment in taiko CLI
     */
    static isContainerEnvironment() {
        return process.env.IS_CONTAINER !== undefined || is_docker_1.default();
    }
}
/**
 * Returns the absolute path of the run output directory.
 */
ScreenCheck.getRunDir = () => path_1.default.join(ScreenCheck.baseDir || "", ScreenCheck.runId || "");
/**
 * Returns the absolute path of the reference run input directory.
 */
ScreenCheck.getRefDir = () => path_1.default.join(ScreenCheck.baseDir || "", ScreenCheck.refRunId || "");
ScreenCheck.defaultFilenameGenerator = async (options, ...args) => `${await ScreenCheck.generateName(ScreenCheck.taiko, options.fullPage, ...args)}.png`;
__decorate([
    cacheInvalidate("detectLatestAutoRunIdIndex"),
    __metadata("design:type", String)
], ScreenCheck, "baseDir", void 0);
__decorate([
    entryPoint(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], ScreenCheck, "init", null);
__decorate([
    exportForPlugin("openBrowser"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Boolean]),
    __metadata("design:returntype", Promise)
], ScreenCheck, "openBrowser", null);
__decorate([
    exportForPlugin("screencheckSetup"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [ScreenCheckSetup]),
    __metadata("design:returntype", ScreenCheckSetup)
], ScreenCheck, "setup", null);
__decorate([
    cached(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Number)
], ScreenCheck, "detectLatestAutoRunIdIndex", null);
__decorate([
    exportForPlugin("isContainerEnvironment"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Boolean)
], ScreenCheck, "isContainerEnvironment", null);
exports.ScreenCheck = ScreenCheck;
exports.default = ScreenCheck;
